# -*- coding: utf-8 -*-
"""
模块: qubo

功能: 校验、调整QUBO模型参数精度，满足真机位宽限制
"""
from kaiwu_community.ising import calculate_ising_matrix_bit_width, adjust_ising_matrix_precision
from kaiwu_community.conversion import qubo_matrix_to_ising_matrix, ising_matrix_to_qubo_matrix


def check_qubo_matrix_bit_width(qubo_matrix, bit_width=8):
    """校验QUBO矩阵元素位宽

    将QUBO矩阵转为伊辛矩阵，通过校验伊辛矩阵的元素位宽来实现对QUBO矩阵的校验

    Args:
        qubo_matrix (np.ndarray): QUBO矩阵

        bit_width (int): 位宽

    Raises:
        ValueError: 当矩阵元素位宽超过指定位宽时抛出异常

    Examples1:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> _matrix = -np.array([[-480., 508., -48.],
        ...                      [ 508., -508., -48.],
        ...                      [ -48., -48., 60.]])
        >>> kw.qubo.check_qubo_matrix_bit_width(_matrix)

    Examples2（缩放后符合要求）:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> _matrix = -np.array([[-512.,  520.,  -48.],
        ...                      [ 520., -520.,  -48.],
        ...                      [ -48.,  -48.,   40.]])
        >>> kw.qubo.check_qubo_matrix_bit_width(_matrix)

    Examples3(缩放后也不符合要求):
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> _matrix = -np.array([[-488.,  516.,  -48.],
        ...                      [ 516., -516.,  -48.],
        ...                      [ -48.,  -48.,   60.]])
        >>> kw.qubo.check_qubo_matrix_bit_width(_matrix)
        Traceback (most recent call last):
        ...
        ValueError: CIM only supports signed 8-bit number
    """
    # qubo模型转ising
    ising_matrix, _ = qubo_matrix_to_ising_matrix(qubo_matrix)
    result = calculate_ising_matrix_bit_width(ising_matrix, bit_width)
    if result.get('precision') > bit_width:
        raise ValueError("CIM only supports signed 8-bit number")


def adjust_qubo_matrix_precision(qubo_matrix, bit_width=8):
    """调整矩阵精度, 通过此接口调整后矩阵可能会有较大的精度损失，比如矩阵有一个数远大于其它数时，调整后矩阵精度损失严重无法使用

    Args:
        qubo_matrix (np.ndarray): 目标矩阵

        bit_width (int): 精度范围，目前只支持8位，有一位是符号位

    Returns:
        np.ndarray: 符合精度要求的QUBO矩阵

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> ori_qubo_mat1 = np.array([[0.89, 0.22, 0.198],
        ...                      [0.22, 0.23, 0.197],
        ...                      [0.198, 0.197, 0.198]])
        >>> qubo_mat1 = kw.qubo.adjust_qubo_matrix_precision(ori_qubo_mat1)
        >>> qubo_mat1
        array([[348., 168., 152.],
               [ -0.,  92., 152.],
               [ -0.,  -0.,  80.]])
        >>> ori_qubo_mat2 = np.array([[0.89, 0.22, 0.198],
        ...                           [0.22, 0.23, 0.197],
        ...                           [0.198, 0.197, 100]])
        >>> qubo_mat2 = kw.qubo.adjust_qubo_matrix_precision(ori_qubo_mat2)
        >>> qubo_mat2  # The solutions obtained by qubo_mat2 and ori_qubo_mat2 matrices are quite different
        array([[  8.,  -0.,  -0.],
               [ -0.,   4.,  -0.],
               [ -0.,  -0., 508.]])
    """
    # qubo矩阵转ising
    ising_matrix, _ = qubo_matrix_to_ising_matrix(qubo_matrix)
    # 校验矩阵精度范围，如何符合要求直接返回
    result = calculate_ising_matrix_bit_width(ising_matrix)
    if result.get('precision') > bit_width:
        _ising_matrix = adjust_ising_matrix_precision(ising_matrix, bit_width)
        # 符合精度要求的qubo矩阵
        qubo_matrix, _ = ising_matrix_to_qubo_matrix(_ising_matrix)
    return qubo_matrix


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
日志
"""
import logging
import sys

# 模块级单例初始化（Python保证模块导入时只执行一次）
_root_logger = logging.getLogger("kaiwu_community")
_root_logger.setLevel(logging.INFO)
_root_logger.propagate = False  # 禁止向上传播

# 统一配置控制台Handler
_console_handler = logging.StreamHandler(sys.stdout)
_console_handler.setLevel(logging.NOTSET)  # 处理所有级别

# 标准化格式
_formatter = logging.Formatter(
    '[%(asctime)s] [%(levelname)-8s] [%(name)s:%(lineno)d] - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S'
)
_console_handler.setFormatter(_formatter)

# 确保只添加一次Handler
if not _root_logger.handlers:
    _root_logger.addHandler(_console_handler)


def set_log_level(level):
    """设置SDK日志输出级别

    SDK默认输出INFO级别以上的日志，可通过此函数修改输出级别

    Args:
        level (str or int): 支持传入 logging.ERROR， logging.INFO、... 、logging.ERROR或字符串ERROR、INFO、... 、ERROR

    Examples:
        >>> import kaiwu_community as kw
        >>> kw.common.set_log_level(level="DEBUG")  # doctest: +SKIP
    """
    level = level.upper() if isinstance(level, str) else level
    _root_logger.setLevel(level)


def set_log_path(path="/tmp/output.log"):
    """设置SDK日志输出文件路径, 需要绝对路径.

    Args:
        path (str): 自定义日志文件输出路径

    Examples:
        >>> import kaiwu_community as kw
        >>> kw.common.set_log_path("/tmp/output.log")  # doctest: +SKIP
    """
    file_handler = logging.FileHandler(path)
    file_handler.setFormatter(_formatter)
    _root_logger.addHandler(file_handler)



"""
基础模块 Optimizer base class
"""
import logging

logger = logging.getLogger(__name__)


class OptimizerBase:
    """
    Ising求解器基类
    """

    def __init__(self):
        self.matrix = None

    def set_matrix(self, ising_matrix):
        """设置矩阵并更新相关内容"""
        if ising_matrix is not None:
            self.matrix = ising_matrix
            self.on_matrix_change()
            logger.debug("Matrix set in the optimizer.")
        if self.matrix is None:
            raise ValueError("Ising matrix must be set")

    def on_matrix_change(self):
        """
        更新矩阵相关信息, 继承OptimizerBase时可以实现。当处理的ising矩阵发生变化时，这个函数的实现会被调用，从而有机会做相应动作
        """

        logger.debug("Method on_matrix_change is not implemented in the class inherited from OptimizerBase!")

    def solve(self, ising_matrix=None):
        """求解"""
        raise NotImplementedError



"""
BinaryExpression
Binary变量构成的表达式
"""

import copy
import math
import numbers
import numpy as np

from kaiwu_community.core._expression import update_constraint
from kaiwu_community.core._expression import Expression, expr_add, expr_neg, expr_mul, expr_pow
from kaiwu_community.core._error import KaiwuError


class BinaryExpression(Expression):
    """QUBO表达式的基础数据结构"""

    def __init__(self, coefficient: dict = None, offset: float = 0, name=""):
        super().__init__(coefficient, offset)
        self.name = name

    def feed(self, feed_dict):
        """为占位符号赋值, 并返回赋值后的新表达式对象

        Args:
            feed_dict(dict): 需要赋值的占位符的值

        Examples:
            >>> import kaiwu_community as kw
            >>> p = kw.core.Placeholder('p')
            >>> a = kw.core.Binary('a')
            >>> y = p * a
            >>> str(y)  # doctest: +NORMALIZE_WHITESPACE
            '(p)*a'
            >>> y= y.feed({'p': 2})
            >>> str(y)  # doctest: +NORMALIZE_WHITESPACE
            '2*a'
        """
        ret = copy.deepcopy(self)
        for expr_vars in ret.coefficient:
            if not isinstance(ret.coefficient[expr_vars], numbers.Number):
                ret.coefficient[expr_vars] = ret.coefficient[expr_vars].feed(feed_dict)
        if not isinstance(ret.offset, numbers.Number):
            ret.offset = ret.offset.feed(feed_dict)
        return ret

    def __repr__(self):
        return self.__str__()

    def __radd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        if isinstance(other, np.ndarray):
            return other.__add__(self)
        if other is None:
            ret = copy.deepcopy(self)
            return ret
        result = BinaryExpression()
        expr_add(self, other, result)
        return result

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __sub__(self, other):
        result = self.__add__(-other)
        return result

    def __neg__(self):
        result = BinaryExpression()
        expr_neg(self, result)
        return result

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):
        if other is None:
            return copy.deepcopy(self)
        result = BinaryExpression()
        expr_mul(self, other, result)
        return result

    def __pow__(self, other):
        result = BinaryExpression()
        expr_pow(self, other, result)
        return result


class Binary(BinaryExpression):
    """二进制变量, 只保存变量名，不继承 QuboExpression"""

    def __init__(self, name: str = ""):
        super().__init__({("b" + name,): 1}, 0, name="b" + name)

    def clear(self):
        """初始化所有属性"""
        self.name = ""
        self.coefficient = {}
        self.offset = 0


class Integer(BinaryExpression):
    """整数变量, 只保存变量名和范围，不继承 QuboExpression"""

    def __init__(self, name: str = "", min_value=0, max_value=127):
        super().__init__()
        self.offset = min_value
        if max_value <= min_value:
            raise KaiwuError("max_value must be larger than min_value")
        num_bits = int(math.log2(max_value - min_value))
        self.coefficient = {}
        for j in range(num_bits):
            self.coefficient[(f"b{name}[{j}]",)] = 2 ** j
        self.coefficient[(f"b{name}[{num_bits}]",)] = max_value - min_value - 2 ** (num_bits) + 1


class Placeholder(BinaryExpression):
    """占位符变量, 只保存变量名, 对决策"""

    def __init__(self, name: str = ""):
        super().__init__()
        self.name = "p" + name
        name = "p" + name
        self.coefficient = {}
        self.offset = _Placeholder({tuple({name}): 1}, 0)

    def get_placeholder_set(self):
        """获取占位符集合"""
        placeholder_set = set()
        for var_tuple in self.coefficient:
            for var in var_tuple:
                placeholder_set.add(var)
        return placeholder_set


class _Placeholder(Expression):
    """占位符的底层实现，实际在QuboExpression的dict结构的参数位置"""

    def feed(self, feed_dict):
        """为占位符赋值"""
        placeholder_value = 0
        for placeholder_vars in self.coefficient:
            p_value = self.coefficient[placeholder_vars]
            for p_var in placeholder_vars:
                p_value *= feed_dict[p_var[1:]]
            placeholder_value += p_value
        placeholder_value += self.offset
        return placeholder_value


def quicksum(qubo_expr_list: list):
    """高性能的QUBO求和器.

    Args:
        qubo_expr_list (QUBO列表): 用于求和的QUBO表达式的列表.

    Returns:
        BinaryExpression: 约束QUBO.

    Examples:
        >>> import kaiwu_community as kw
        >>> qubo_list = [kw.core.Binary("b"+str(i)) for i in range(10)] # Variables are also QUBO
        >>> output = kw.core.quicksum(qubo_list)
        >>> str(output)
        'b0+b1+b2+b3+b4+b5+b6+b7+b8+b9'
    """
    qsum = BinaryExpression()
    for single_q in qubo_expr_list:
        if isinstance(single_q, numbers.Number):
            qsum.offset += single_q
            continue
        if not isinstance(single_q, dict):
            raise KaiwuError("qubo_expr_list should be a list of QUBO Expression")

        update_constraint(single_q, qsum)

        qsum.offset += single_q.offset
        for ele in single_q.coefficient.keys():
            if ele in qsum.coefficient:
                qsum.coefficient[ele] += single_q.coefficient[ele]
            else:
                qsum.coefficient[ele] = single_q.coefficient[ele]
            if qsum.coefficient[ele] == 0:
                qsum.coefficient.pop(ele)
    return qsum


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
模块: Ising

功能: 提供Ising模型相关函数
"""
from kaiwu_community.ising._ising import IsingModel, Spin, IsingExpression
from kaiwu_community.ising._precision import calculate_ising_matrix_bit_width, adjust_ising_matrix_precision

__all__ = [
    "IsingModel",
    "IsingExpression",
    "calculate_ising_matrix_bit_width",
    "adjust_ising_matrix_precision",
    "Spin"
]



# -*- coding: utf-8 -*-
"""
模块: core.PMConstraint

功能: 生成基于penalty method的约束项
"""
import logging
from kaiwu_community.core._binary_expression import Integer
from kaiwu_community.core._get_val import get_val
from kaiwu_community.core._constraint import ConstraintDefinition
logger = logging.getLogger(__name__)


class PenaltyMethodConstraint:
    """有约束转无约束的penalty method方法

    Args:
        expr (Expression): 编译后的约束项表达式

        penalty (float): 约束项惩罚系数

    """

    def __init__(self, expr, penalty=1, parent_model=None):

        self.constraint_expr = expr
        self.previous_penalty = 1
        self.penalty = penalty
        self.pre_constr_val = 0
        self.current_value = 0
        self._parent_model = parent_model

    @classmethod
    def from_constraint_definition(cls, name, constraint: ConstraintDefinition, parent_model):

        """Prepare QUBO expression for the given constraint, automatically determining slack variables if needed.

        Args:
         name: Name of the constraint.

         constraint: The relation constraint to process.

         parent_model: the model it belongs to.

        """

        if constraint.relation == '==':
            # 等式约束直接平方处理
            expr = constraint.left_operand - constraint.expected_value
        else:
            # 处理不等式约束的方向
            diff_qubo = _adjust_inequality_direction(constraint)
            slack_expr = _create_slack_variable(name, diff_qubo, constraint.relation)
            # 构建最终的约束表达式（等式平方形式）
            expr = diff_qubo + slack_expr - constraint.expected_value
            expr = expr ** 2

        logger.debug("Constraint expression: %s", expr)

        return PenaltyMethodConstraint(expr, constraint.default_penalty, parent_model)

    def set_penalty(self, penalty):
        """ 设置惩罚系数"""
        if penalty is None:
            penalty = 1
        self.previous_penalty = self.penalty
        self.penalty = penalty
        if self._parent_model:
            self._parent_model.invalidate_made_state()
        logger.debug("Penalty: %s Constraint expression: %s", self.penalty, self.constraint_expr)

    def penalize_more(self):
        """增加惩罚系数"""
        self.set_penalty(self.penalty * 2)

    def penalize_less(self):
        """降低惩罚系数"""
        if self.previous_penalty is None:
            self.previous_penalty = 0
        self.set_penalty((self.previous_penalty + self.penalty) / 2)

    def __str__(self):
        return f"penalty={self.penalty}, constraint_expr={self.constraint_expr}"

    def __repr__(self) -> str:
        return self.__str__()

    def is_satisfied(self, solution_dict):
        """验证约束满足情况"""
        self.current_value = float(get_val(self.constraint_expr, solution_dict))
        return abs(self.current_value) < 1e-5


def _find_min_interval(diff_qubo):
    """找到系数间的最小正间隔"""
    coefficients = sorted(diff_qubo.coefficient.values())
    # 添加0以处理边界情况
    coefficients.append(0)
    # 计算相邻系数的最小正差值
    min_diff = min(
        abs(coefficients[i + 1] - coefficients[i])
        for i in range(len(coefficients) - 1)
        if coefficients[i + 1] != coefficients[i]
    )
    return min_diff


def _calculate_slack_range(diff_qubo):
    """计算松弛变量的最大取值范围"""
    # 累加所有负系数（考虑最坏情况）
    negative_sum = sum(-v for v in diff_qubo.coefficient.values() if v < 0)
    # 计算初始范围并确保非负
    slack_range = max(0, negative_sum - diff_qubo.offset)
    return slack_range


def _create_slack_variable(name, diff_qubo, relation):
    """自动创建松弛变量表达式"""
    # 确定松弛变量的最小值
    slack_min = 0 if relation in ['>=', '<='] else 1

    # 计算松弛变量的取值范围
    slack_range = _calculate_slack_range(diff_qubo)
    if slack_range == 0:
        raise ValueError("Slack range cannot be zero for non-equality constraint")

    # 确定离散化精度
    min_diff = _find_min_interval(diff_qubo)
    precision_steps = int(round(slack_range / min_diff))

    # 创建整数变量并线性缩放
    slack_var = Integer(f"_slack_{name}", slack_min, precision_steps + slack_min)
    return slack_var * (slack_range / precision_steps)


def _adjust_inequality_direction(constraint):
    """根据不等式方向调整QUBO表达式符号"""
    if constraint.relation in ['>', '>=']:
        return -constraint.left_operand
    return constraint.left_operand



# -*- coding: utf-8 -*-
"""
模块: qubo

功能: QUBO矩阵和Ising矩阵相关转化
"""
from decimal import Decimal
import numpy as np


def ising_matrix_to_qubo_matrix(ising_mat, remove_linear_bit=True):
    """Ising矩阵转QUBO矩阵

    Args:
        ising_mat (np.ndarray): Ising矩阵

        remove_linear_bit (bool): QUBO转Ising时会增加一个辅助变量表示线性项。是否移除最后一个自旋变量。默认为True。

    Returns:
        tuple: QUBO矩阵和bias

            - qubo_mat (np.ndarray): QUBO矩阵
            - bias (float): QUBO与Ising相差的常数项

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> matrix = -np.array([[ 0. ,  1. ,  0. ,  1. ,  1. ],
        ...                     [ 1. ,  0. ,  0. ,  1.,   1. ],
        ...                     [ 0. ,  0. ,  0. ,  1.,   1. ],
        ...                     [ 1. ,  1.,   1. ,  0. ,  1. ],
        ...                     [ 1. ,  1.,   1. ,  1. ,  0. ]])
        >>> _qubo_mat, _ = kw.conversion.ising_matrix_to_qubo_matrix(matrix)
        >>> _qubo_mat
        array([[-4.,  8.,  0.,  8.],
               [-0., -4.,  0.,  8.],
               [-0., -0., -0.,  8.],
               [-0., -0., -0., -8.]])
    """
    # 计算前转Decimal以防精度损失
    ising_mat = np.vectorize(str)(ising_mat)
    ising_mat = np.vectorize(Decimal)(ising_mat)

    ising_mat = (ising_mat + ising_mat.T) / 2
    ising_mat_linear = None
    if remove_linear_bit:
        ising_mat_linear = ising_mat[-1, :-1] * 2
        ising_mat = ising_mat[:-1, :-1]
    qubo_mat = ising_mat * 4
    diag_vec = - 4 * np.sum(ising_mat, axis=0)
    bias = np.sum(ising_mat)
    if remove_linear_bit:
        bias -= np.sum(ising_mat_linear)
        diag_vec += 2 * ising_mat_linear
    qubo_mat = np.triu(qubo_mat * 2)
    np.fill_diagonal(qubo_mat, diag_vec)
    qubo_mat = np.array(qubo_mat, dtype=np.float64)
    return -qubo_mat, -float(bias)


def qubo_matrix_to_ising_matrix(qubo_mat):
    """QUBO矩阵转Ising矩阵

    Args:
        qubo_mat (np.ndarray): QUBO矩阵

    Returns:
        tuple: Ising矩阵和bias
            - ising_mat (np.ndarray): Ising矩阵
            - bias (float): QUBO与Ising相差的常数项

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> matrix = -np.array([[-4.,  8.,  0.,  8.],
        ...                     [-0., -4.,  0.,  8.],
        ...                     [-0., -0., -0.,  8.],
        ...                     [-0., -0., -0., -8.]])
        >>> _ising_mat, _ = kw.conversion.qubo_matrix_to_ising_matrix(matrix)
        >>> _ising_mat
        array([[-0.,  1., -0.,  1.,  1.],
               [ 1., -0., -0.,  1.,  1.],
               [-0., -0., -0.,  1.,  1.],
               [ 1.,  1.,  1., -0.,  1.],
               [ 1.,  1.,  1.,  1., -0.]])
    """
    # 计算前转Decimal以防精度损失
    qubo_mat = np.vectorize(str)(qubo_mat)
    qubo_mat = np.vectorize(Decimal)(qubo_mat)

    ising_size = qubo_mat.shape[0] + 1
    ising_mat = np.zeros((ising_size, ising_size))
    qubo_mat = (qubo_mat + qubo_mat.T) / 8
    qubo_div_4_diagoal = np.diagonal(qubo_mat).copy()
    qubo_div_4 = qubo_mat
    np.fill_diagonal(qubo_div_4, 0)
    ising_mat[:-1, :-1] = qubo_div_4
    ising_mat[-1, :-1] = np.sum(qubo_div_4, axis=0) + qubo_div_4_diagoal
    ising_mat[:-1, -1] = np.sum(qubo_div_4, axis=0) + qubo_div_4_diagoal
    bias = np.sum(qubo_div_4) + np.sum(qubo_div_4_diagoal) * 2
    ising_mat = np.array(ising_mat, dtype=np.float64)
    return -ising_mat, float(bias)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
通用工具集合
"""

from kaiwu_community.common._logger import set_log_level, set_log_path
from kaiwu_community.common._util import hamiltonian, check_symmetric

__all__ = [
    "hamiltonian",
    "check_symmetric",
    "set_log_level",
    "set_log_path",
]



"""
模块: Model

功能: Binary Model基础类
"""
import collections.abc
import copy
from typing import Literal
import logging
import numpy as np
from kaiwu_community.core._penalty_method_constraint import PenaltyMethodConstraint
from kaiwu_community.core._constraint import get_min_penalty_from_min_diff, get_soft_penalty
from kaiwu_community.core._get_val import get_val
from kaiwu_community.core._error import KaiwuError

logger = logging.getLogger(__name__)


class BinaryModel:
    """二值模型类

    Args:
        objective (BinaryExpression, optional): 目标函数. 默认为None
    """

    def __init__(self, objective=None):
        self.objective = copy.deepcopy(objective)
        self.hard_constraints = {}
        self.soft_constraints = {}
        self.hard_constraints_made = {}
        self.soft_constraints_made = {}
        self.compiled = False
        self.constraint_handler = PenaltyMethodConstraint

    def set_constraint_handler(self, constraint_handler):
        """设置约束项无约束化方法

        Args:
            constraint_handler: 设置约束项无约束化表示方法类
        """
        self.constraint_handler = constraint_handler

    def set_objective(self, objective):
        """设置目标函数

        Args:
            objective (BinaryExpression): 目标函数表达式
        """
        if self.objective is not None:
            raise KaiwuError("Object already exists, please reinitialize")
        self.objective = copy.deepcopy(objective)

    def add_constraint(self, constraint_in, name=None, constr_type: Literal["soft", "hard"] = "hard", penalty=None):
        """添加约束项（支持单个或多个约束）

        Args:
            constraint_in (ConstraintDefinition or iterable): 约束表达式或其可迭代对象

            name (str or list, optional): 约束名称或名称列表，默认自动命名

            constr_type (str, optional): 约束类型，可以设置为"soft"或"hard"，默认为"hard"

            penalty (float): 缺省惩罚系数
        """

        if constr_type not in ["hard", "soft"]:
            raise KaiwuError(f"No such type {constr_type}")

        # Handle multiple constraints
        if isinstance(constraint_in, (list, tuple, np.ndarray)) or (
            isinstance(constraint_in, collections.abc.Iterable) and not isinstance(constraint_in, (str, bytes))
        ):
            # If name is a list/tuple, use as names; else auto-generate
            names = name if isinstance(name, (list, tuple, np.ndarray)) else None
            for idx, constraint in enumerate(constraint_in):
                item_name = (
                    names[idx] if names is not None and idx < len(names)
                    else f"constraint{len(self.hard_constraints) + len(self.soft_constraints) + idx}"
                )
                self.add_constraint(constraint, item_name, constr_type, penalty)
            return

        # Single constraint
        if name is None:
            name = f"constraint{len(self.hard_constraints) + len(self.soft_constraints)}"
        if name in self.hard_constraints or name in self.soft_constraints:
            logger.warning("Constraint %s is already added. The original one will be replaced.", name)
        constraint_in.default_penalty = penalty
        if constr_type == "soft":
            self.soft_constraints[name] = constraint_in
        else:
            self.hard_constraints[name] = constraint_in

    def get_value(self, solution_dict):
        """根据结果字典将变量值带入qubo变量.

        Args:
            solution_dict (dict): 由get_sol_dict生成的结果字典。

        Returns:
            float: 带入qubo后所得的值
        """
        return get_val(self.objective, solution_dict)

    def verify_constraint(self, solution_dict, constr_type: Literal["soft", "hard"] = "hard"):
        """确认约束是否满足

        Args:
            solution_dict (dict): QUBO模型解字典

            constr_type(str, optional): 约束类型，可以设置为"soft"或"hard"，默认为"hard"

        Returns:
            tuple: 约束满足信息
                - int: 不满足的约束个数
                - dict: 包含约束值的字典
        """
        if constr_type not in ["soft", "hard"]:
            raise KaiwuError(f"No such type {constr_type}")
        unsatisfied_count = 0
        result_dict = {}
        temp_constraint_dicts = self.hard_constraints
        if constr_type == "soft":
            temp_constraint_dicts = self.soft_constraints

        for name, constr_info in temp_constraint_dicts.items():
            if not constr_info.is_satisfied(solution_dict):
                unsatisfied_count += 1
            result_dict[name] = get_val(constr_info.left_operand, solution_dict)

        logger.debug("unsatisfied %s constraints count = %s | objective = %s",
                     constr_type, unsatisfied_count, self.get_value(solution_dict))

        return unsatisfied_count, result_dict

    def initialize_penalties(self):
        """ 自动初始化所有的惩罚系数"""
        positive_delta = {}
        negative_delta = {}

        if self.objective is not None:
            negative_delta, positive_delta = self.objective.get_max_deltas()

        for _, constraint_info in self.hard_constraints_made.items():
            constraint_info.set_penalty(
                get_min_penalty_from_min_diff(constraint_info.constraint_expr, negative_delta, positive_delta))
        for _, constraint_info in self.soft_constraints_made.items():
            constraint_info.set_penalty(get_soft_penalty(self.objective, constraint_info.constraint_expr))

    def get_constraints_expr_list(self):
        """获取当前所有的constraint。

        Returns:
            List of all constraints.

        """

        constraint_expr_list = []
        for _, constraint_info in self.hard_constraints_made.items():
            constraint_expr_list.append(constraint_info.penalty * constraint_info.constraint_expr)
        for _, constraint_info in self.soft_constraints_made.items():
            constraint_expr_list.append(constraint_info.penalty * constraint_info.constraint_expr)
        return constraint_expr_list

    def compile_constraints(self):
        """ 按照不同的风格转换约束项为Expression
        对于不等式约束，目前支持的是罚函数方式，
        """
        if self.compiled:
            return
        if self.constraint_handler is None:
            raise KaiwuError("Please set constraint handler first!")

        # 整理约束项
        for name, constraint in self.hard_constraints.items():
            self.hard_constraints_made[name] = (
                self.constraint_handler.from_constraint_definition(name, constraint, self))
        for name, constraint in self.soft_constraints.items():
            self.soft_constraints_made[name] = (
                self.constraint_handler.from_constraint_definition(name, constraint, self))
        self.compiled = True



# -*- coding: utf-8 -*-
"""
模块: cim

功能: ising模型及其转化
"""
from kaiwu_community.core import Expression


def _dict_variables(var_dict):
    var_list = [0] * len(var_dict)
    for var in var_dict:
        var_list[var_dict[var]] = var[1:]
    return str(tuple(var_list)).replace("\'", "")[1:-1]


class IsingModel(dict):
    """ising模型"""

    def __init__(self, variables, ising_matrix, bias):
        super().__init__()
        self.variables = variables
        self.matrix = ising_matrix
        self.bias = bias

    def __repr__(self):
        return f"{self.__class__.__name__}({vars(self)!r})"

    def __str__(self):
        """返回ising模型的细节信息.

        Returns:
            序列化后的模型(字符串形式）
        """
        print_data = ""
        print_data += "CIM Ising Details:\n"
        print_data += "  CIM Ising Matrix:\n"
        print_data += "    " + str(self.matrix).replace("\n", "\n    ") + "\n"
        print_data += "  CIM Ising Bias: " + str(self.bias) + "\n"
        print_data += "  CIM Ising Variables" + ": " + \
                      _dict_variables(self.variables) + "\n"

        return print_data

    def get_variables(self):
        """获取模型中的变量"""
        return dict(zip((key[1:] for key in self.variables), self.variables.values()))

    def get_matrix(self):
        """获取Ising矩阵"""
        return self.matrix

    def get_bias(self):
        """获取QUBO转化时得到的常数偏置"""
        return self.bias


class IsingExpression(Expression):
    """Ising 表达式基类，直接继承 Expression，保留扩展点。"""

    def __init__(self, variables=None, quadratic=None, linear=None, bias=0):
        if quadratic is None:
            quadratic = {}
        if linear is None:
            linear = {}
        super().__init__({**quadratic, **linear}, bias)
        self.variables = variables
        self.quadratics = quadratic
        self.linear = linear
        self.bias = bias


class Spin(IsingExpression):
    """自旋变量, 可能的取值只有-1,1.

    Args:
        name (str): 变量的唯一标识.

    Returns:
        dict: 名称为name的自旋变量.

    Examples:
        >>> import kaiwu_community as kw
        >>> s = kw.ising.Spin("s")
        >>> s
        2*s-1
    """

    def __init__(self, name: str = ""):
        name = "s" + name
        super().__init__(linear={tuple({name}): 2}, bias=-1)
        self.name = name


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
模块: qubo

功能: QUBO模型和Ising模型转换
"""
import numpy as np
from kaiwu_community.ising import IsingModel, IsingExpression



def _to_ising(qubo_expr):
    """QUBO表达式转化为Ising模型"""
    variables_set = set()
    quadratic = {}
    linear = {}
    bias = 0
    for key in qubo_expr.coefficient:
        variables_set.add(key[0])
        if len(key) == 2:
            variables_set.add(key[1])
            coefficient = qubo_expr.coefficient[key] / 4
            quadratic[key] = coefficient
            if key[0] in linear:
                linear[key[0]] += coefficient
            else:
                linear[key[0]] = coefficient
            if key[1] in linear:
                linear[key[1]] += coefficient
            else:
                linear[key[1]] = coefficient
            bias += coefficient
        else:
            coefficient = qubo_expr.coefficient[key] / 2
            if key[0] in linear:
                linear[key[0]] += coefficient
            else:
                linear[key[0]] = coefficient
            bias += coefficient
    bias += qubo_expr.offset

    variables_set = list(variables_set)
    variables_set.sort()
    variables_set = tuple(variables_set)

    return IsingExpression(variables_set, quadratic, linear, bias)


def qubo_model_to_ising_model(qubo_model):
    """QUBO转CIM Ising模型.

    Args:
        qubo_model (QuboModel): QUBO Model.

    Returns:
        CimIsing: CIM Ising模型.

    Examples:
        >>> import kaiwu_community as kw
        >>> b1, b2 = kw.core.Binary("b1"), kw.core.Binary("b2")
        >>> q = b1 + b2 + b1*b2
        >>> q_model = kw.qubo.QuboModel(q)
        >>> ci = kw.conversion.qubo_model_to_ising_model(q_model)
        >>> print(str(ci))
        CIM Ising Details:
          CIM Ising Matrix:
            [[-0.    -0.125 -0.375]
             [-0.125 -0.    -0.375]
             [-0.375 -0.375 -0.   ]]
          CIM Ising Bias: 1.25
          CIM Ising Variables: b1, b2, __spin__
        <BLANKLINE>
    """
    qubo_model.compile_constraints()
    qubo_expr = qubo_model.make()
    ising_expression = _to_ising(qubo_expr)
    variable_index = dict(zip(ising_expression.variables, range(len(ising_expression.variables))))
    variable_index["a__spin__"] = len(ising_expression.variables)
    cim_matrix = np.zeros((len(ising_expression.variables) + 1, len(ising_expression.variables) + 1))
    for key in ising_expression.quadratics:
        cim_matrix[variable_index[key[0]], variable_index[key[1]]] = ising_expression.quadratics[key]

    for key in ising_expression.linear:
        cim_matrix[variable_index[key], len(ising_expression.variables)] = ising_expression.linear[key]

    cim_matrix = cim_matrix + cim_matrix.T

    return IsingModel(variable_index, -0.5 * cim_matrix, ising_expression.bias)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
模块: qubo

功能: QUBO相关的错误处理
"""
from kaiwu_community.core import KaiwuError


class QuboError(KaiwuError):
    """Exceptions in qubo module."""

    def __init__(self, error_info):
        self.error_info = error_info

    def __str__(self):
        return self.error_info



"""
转换工具集合
"""
from kaiwu_community.conversion._matrix_converter import ising_matrix_to_qubo_matrix, qubo_matrix_to_ising_matrix
from kaiwu_community.conversion._model_converter import qubo_model_to_ising_model

__all__ = [
    "ising_matrix_to_qubo_matrix", "qubo_matrix_to_ising_matrix",
    "qubo_model_to_ising_model"

]



# encoding=utf8
"""
SolverBase
Provide some basic implementation of solver and its interface
"""
import logging
import abc
from functools import wraps

logger = logging.getLogger(__name__)


class SolverBaseMeta(abc.ABCMeta):
    """Solver自定义元类"""

    def __new__(mcs, name, bases, namespace, **kwargs):
        if name != 'Base' and 'solve_qubo' in namespace:
            original_execute = namespace['solve_qubo']

            @wraps(original_execute)
            def wrapped_execute(self, *args, **kwargs):
                result = original_execute(self, *args, **kwargs)
                if result[0] is None:
                    logger.warning("No solution found!")
                return result

            namespace['solve_qubo'] = wrapped_execute
        return super().__new__(mcs, name, bases, namespace, **kwargs)


class SolverBase(metaclass=SolverBaseMeta):
    """Solver基类

    Args:
        optimizer (OptimizerBase): Ising求解器
    """

    def __init__(self, optimizer):
        self._optimizer = optimizer

    @abc.abstractmethod
    def solve_qubo(self, qubo_model):
        """求解QUBO

        Args:
            qubo_model (QuboModel): QUBO模型

        Returns:
            tuple: 元组，包含求解结果信息
                - dict: 解字典
                - float: QUBO值
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
时间: 2021-07-07
作者: wangyong@boseq.com
"""
from typing import Union, Tuple, List
import numpy as np
from kaiwu_community.core._binary_expression import quicksum
from kaiwu_community.core._expression import Expression


def dot(mat_left, mat_right):
    """矩阵乘法

    Args:
        mat_left (numpy.array): 矩阵1

        mat_right (numpy.array): 矩阵2

    Raises:
        ValueError: 两个输入都必须是np.ndarray
        ValueError: 两个输入的维度必须匹配

    Returns:
        np.ndarray: 乘积矩阵
    """
    # 检查输入是否为NumPy数组
    if not isinstance(mat_left, np.ndarray) or not isinstance(mat_right, np.ndarray):
        raise ValueError("Both inputs must be NumPy arrays")
    left_is_vector = len(mat_left.shape) == 1
    right_is_vector = len(mat_right.shape) == 1
    # 为了方便使用矩阵运算的写法进行运算，根据位置把向量补全成为矩阵
    if left_is_vector:
        mat_left = mat_left[np.newaxis, :]
    if right_is_vector:
        mat_right = mat_right[:, np.newaxis]
    mat_right = mat_right.swapaxes(-1, -2)

    # 获取数组的形状
    mat_left_shape = mat_left.shape
    mat_right_shape = mat_right.shape

    # 确保两个输入的最后一个维度相同，或者其中一个是1
    if mat_left_shape[-1] != mat_right_shape[-1]:
        raise ValueError("The last dimension of A must be equal to the"
                         " second last dimension of B for dot product")

    # 初始化结果数组
    result_shape = mat_left_shape[:-1] + mat_right_shape[:-1]
    result = BinaryExpressionNDArray(result_shape, dtype=Expression)

    # 使用嵌套循环和np.sum来实现多维点积
    for a_idx in np.ndindex(mat_left_shape[:-1]):
        for b_idx in np.ndindex(mat_right_shape[:-1]):
            result[a_idx + b_idx] = quicksum((mat_left[a_idx] * mat_right[b_idx]).tolist())

    if right_is_vector:
        result = result.squeeze(-1)
    if left_is_vector:
        result = result[0]
    return result


def _is_less(qubo_left, qubo_right):
    return qubo_left < qubo_right


def _is_less_equal(qubo_left, qubo_right):
    return qubo_left <= qubo_right


def _is_greater(qubo_left, qubo_right):
    return qubo_left > qubo_right


def _is_greater_equal(qubo_left, qubo_right):
    return qubo_left >= qubo_right


def _is_equal(qubo_left, qubo_right):
    return qubo_left == qubo_right


class BinaryExpressionNDArray(np.ndarray):
    """基于 np.ndarray 的QUBO容器.
    该容器支持各种 numpy 原生的向量化运算
    """

    def __lt__(self, other):
        return np.vectorize(_is_less)(self, other)

    def __le__(self, other):
        return np.vectorize(_is_less_equal)(self, other)

    def __gt__(self, other):
        return np.vectorize(_is_greater)(self, other)

    def __ge__(self, other):
        return np.vectorize(_is_greater_equal)(self, other)

    def __eq__(self, other):
        return np.vectorize(_is_equal)(self, other)

    def __matmul__(self, other):
        return self.dot(other)

    def dot(self, b, out=None):
        """使用quicksum的矩阵乘法

        Args:
            b (BinaryExpressionNDArray): 另一个矩阵

            out: 可选输出数组，用于存储结果。需与预期输出形状一致。

        Returns:
            BinaryExpressionNDArray: 乘积
        """
        if out is not None:
            idx = (slice(None),) * len(out.shape)
            out[idx] = dot(self, b)
            return out
        return dot(self, b)

    # pylint: disable=W0613, R0917
    def sum(self, axis=None, dtype=None, out=None, keepdims=False, initial=0, where=True):
        """使用quicksum的求和方法

        Args:
            axis: 指定求和的轴（维度）。默认为 None，表示对所有元素求和；若为整数或元组，则沿指定轴求和。

            dtype: 指定输出数据类型。若未提供，则默认使用输入数组的 dtype，但整数类型可能提升为平台整数精度。暂不支持。

            out: 可选输出数组，用于存储结果。需与预期输出形状一致。

            keepdims: 布尔值。若为 True，则保留被求和的轴作为长度为1的维度。暂不支持 。

            initial: 求和的初始值（标量），默认为0。暂不支持。

            where: 布尔数组，指定哪些元素参与求和（NumPy 1.20+支持）。暂不支持。

        Returns:
            BinaryExpressionNDArray: 乘积
        """
        # 检查输入是否为NumPy数组
        if not isinstance(self, np.ndarray):
            raise ValueError("Input must be a NumPy array")
        # 如果没有指定轴，则对整个数组求和
        if axis is None:
            # 将数组展平成一个列表并使用quicksum求和
            return quicksum(self.flatten().tolist())
        # 指定轴的情况
        if axis < 0:
            axis += self.ndim  # 转换负轴索引为正轴索引

        # 初始化结果数组
        new_shape = self.shape[:axis] + self.shape[axis + 1:]
        result = BinaryExpressionNDArray(new_shape, dtype=Expression)

        # 对指定轴进行求和
        for idx in np.ndindex(new_shape):
            # 获取当前轴的子数组
            sub_array = self[idx[:axis] + (slice(None),) + idx[axis:]]
            result[idx] = quicksum(sub_array.flatten().tolist())

        if out is not None:
            idx = (slice(None),) * len(out.shape)
            out[idx] = result
            return out
        if not result.shape:
            result = result.tolist()
        return result


def ndarray(shape: Union[int, Tuple[int, ...], List[int]], name, var_func, var_func_param=None):
    """基于 np.ndarray 的QUBO容器.
    该容器支持各种 numpy 原生的向量化运算

    Args:
        shape (Union[int, Tuple[int, ...]]): 形状

        name (str): 生成的变量的标识符.

        var_func (class for func): 用于生成元素的方法或类. 第一个参数必须是name

        var_func_param (tuple): var_func除了name以外的参数 

    Returns:
        np.ndarray: 多维容器.

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> A = kw.core.ndarray((2,3,4), "A", kw.core.Binary)
        >>> A
        BinaryExpressionNDArray([[[A[0][0][0], A[0][0][1], A[0][0][2],
                                   A[0][0][3]],
                                  [A[0][1][0], A[0][1][1], A[0][1][2],
                                   A[0][1][3]],
                                  [A[0][2][0], A[0][2][1], A[0][2][2],
                                   A[0][2][3]]],
        <BLANKLINE>
                                 [[A[1][0][0], A[1][0][1], A[1][0][2],
                                   A[1][0][3]],
                                  [A[1][1][0], A[1][1][1], A[1][1][2],
                                   A[1][1][3]],
                                  [A[1][2][0], A[1][2][1], A[1][2][2],
                                   A[1][2][3]]]], dtype=object)
        >>> A[1,2]
        BinaryExpressionNDArray([A[1][2][0], A[1][2][1], A[1][2][2], A[1][2][3]],
                                dtype=object)
        >>> A[:, [0,2]]
        BinaryExpressionNDArray([[[A[0][0][0], A[0][0][1], A[0][0][2],
                                   A[0][0][3]],
                                  [A[0][2][0], A[0][2][1], A[0][2][2],
                                   A[0][2][3]]],
        <BLANKLINE>
                                 [[A[1][0][0], A[1][0][1], A[1][0][2],
                                   A[1][0][3]],
                                  [A[1][2][0], A[1][2][1], A[1][2][2],
                                   A[1][2][3]]]], dtype=object)
        >>> B = kw.core.ndarray(3, "B", kw.core.Binary)
        >>> B
        BinaryExpressionNDArray([B[0], B[1], B[2]], dtype=object)
        >>> C = kw.core.ndarray([3,3], "C", kw.core.Binary)
        >>> C
        BinaryExpressionNDArray([[C[0][0], C[0][1], C[0][2]],
                                 [C[1][0], C[1][1], C[1][2]],
                                 [C[2][0], C[2][1], C[2][2]]], dtype=object)
        >>> D = 2 * B.dot(C) + 2
        >>> str(D[0])
        '2*B[0]*C[0][0]+2*B[1]*C[1][0]+2*B[2]*C[2][0]+2'
        >>> E = B.sum()
        >>> str(E)
        'B[0]+B[1]+B[2]'
        >>> F = np.diag(C)
        >>> F
        BinaryExpressionNDArray([C[0][0], C[1][1], C[2][2]], dtype=object)
    """
    if isinstance(shape, int):
        shape = (shape,)
    if isinstance(shape, list):
        shape = tuple(shape)
    if 0 in shape:
        raise ValueError("The argument shape cannot contain 0.")
    if var_func_param is None:
        var_func_param = ()

    arr = BinaryExpressionNDArray(shape, dtype=Expression)
    for idx in np.ndindex(shape):
        item_str = name
        for i, idx_i in enumerate(idx):
            len_str_k = len(str(shape[i] - 1))
            item_str += f"[{str(idx_i).zfill(len_str_k)}]"
        arr[idx] = var_func(item_str, *var_func_param)
    return arr


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
模块: qubo

功能: QUBO及Ising表达式运算类
"""
import copy
import numbers
from kaiwu_community.core._error import KaiwuError
from kaiwu_community.core._constraint import ConstraintDefinition

QUBO_MAINTAINED_KEY = ["hard_constraint", "soft_constraint", "hobo_var_dict", "hobo_constraint"]


def is_zero(qubo_expr):
    """QUBO表达式为0"""
    if isinstance(qubo_expr, numbers.Number):
        return qubo_expr == 0
    return len(qubo_expr.coefficient) == 0 and qubo_expr.offset == 0


def update_constraint(qubo_origin, qubo_result):
    """将qexp的约束以及降阶约束信息更新到q"""
    for key in QUBO_MAINTAINED_KEY:
        if key in qubo_origin:
            if key in qubo_result:
                qubo_result[key].update(qubo_origin[key])
            else:
                qubo_result[key] = qubo_origin[key]


def expr_add(expr_left, expr_right, expr_result):
    """通用二次表达式相加"""
    if isinstance(expr_right, numbers.Number):
        expr_result.coefficient = expr_left.coefficient.copy()
        expr_result.offset = expr_left.offset + expr_right
        update_constraint(expr_left, expr_result)
    else:
        expr_result.offset = expr_left.offset + expr_right.offset
        if len(expr_right.coefficient) < len(expr_left.coefficient):
            expr_result.coefficient = expr_left.coefficient.copy()
            expr_other = expr_right
        else:
            expr_result.coefficient = expr_right.coefficient.copy()
            expr_other = expr_left

        for key in expr_other.coefficient:
            if key in expr_result.coefficient:
                expr_result.coefficient[key] += expr_other.coefficient[key]
                if expr_result.coefficient[key] == 0:
                    expr_result.coefficient.pop(key)
            else:
                expr_result.coefficient[key] = expr_other.coefficient[key]


def expr_neg(expr_origin, expr_result):
    """通用二次表达式取负"""
    for key in expr_origin.coefficient:
        expr_result.coefficient[key] = -expr_origin.coefficient[key]
    expr_result.offset = -expr_origin.offset
    update_constraint(expr_origin, expr_result)


def _expr_dicts_mul(expr_left, expr_right, expr_result):
    for lkey in expr_left.coefficient:
        for rkey in expr_right.coefficient:
            key = list(set(lkey + rkey))
            key.sort()
            key = tuple(key)
            if key in expr_result.coefficient:
                expr_result.coefficient[key] += expr_left.coefficient[lkey] * \
                                                expr_right.coefficient[rkey]
                if expr_result.coefficient[key] == 0:
                    expr_result.coefficient.pop(key)
            else:
                expr_result.coefficient[key] = expr_left.coefficient[lkey] * \
                                               expr_right.coefficient[rkey]
        if not is_zero(expr_right.offset):
            if lkey in expr_result.coefficient:
                expr_result.coefficient[lkey] += expr_left.coefficient[lkey] * \
                                                 expr_right.offset
            else:
                expr_result.coefficient[lkey] = expr_left.coefficient[lkey] * \
                                                expr_right.offset
    if not is_zero(expr_left.offset):
        for rkey in expr_right.coefficient:
            if rkey in expr_result.coefficient:
                expr_result.coefficient[rkey] += expr_right.coefficient[rkey] * \
                                                 expr_left.offset
            else:
                expr_result.coefficient[rkey] = expr_right.coefficient[rkey] * \
                                                expr_left.offset
        expr_result.offset += expr_left.offset * expr_right.offset


def expr_mul(expr_left, expr_right, expr_result):
    """通用二次表达式相乘"""
    if isinstance(expr_right, numbers.Number):
        if not is_zero(expr_right):
            for lkey in expr_left.coefficient:
                expr_result.coefficient[lkey] = expr_left.coefficient[lkey] * expr_right
            expr_result.offset = expr_left.offset * expr_right
        update_constraint(expr_left, expr_result)
    else:
        _expr_dicts_mul(expr_left, expr_right, expr_result)


def expr_pow(expr_left, expr_right, expr_result):
    """通用二次表达式乘方，要求expr_right只能为1或者2"""
    if expr_right == 1:
        expr_result.offset = expr_left.offset
        expr_result.coefficient = expr_left.coefficient.copy()
    elif expr_right == 2:
        expr_mul(expr_left, expr_left, expr_result)
    else:
        if len(expr_left.coefficient) == 1 and \
                (list(expr_left.coefficient.keys())[0][0][0] == "b"):
            expr_result.offset = expr_left.offset
            expr_result.coefficient = expr_left.coefficient.copy()
        else:
            raise KaiwuError("Items higher than quadratic.")
    update_constraint(expr_left, expr_result)


def _check_unit(checked_str):
    if len(checked_str) > 2:
        if checked_str[:2] == "1*":
            checked_str = checked_str[2:]
        elif checked_str[:3] == "-1*":
            checked_str = "-" + checked_str[3:]
    return checked_str


class Expression(dict):
    """QUBO/Ising 通用表达式基类（提供默认二次表达式实现）"""

    def __init__(self, coefficient: dict = None, offset: float = 0):
        super().__init__()
        if coefficient is None:
            self.coefficient = {}
        else:
            self.coefficient = coefficient

        self.offset = offset

    def clear(self) -> None:
        """初始化所有属性"""
        self.coefficient = {}
        self.offset = 0

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        expression_str = ""
        for p_key in self.coefficient:
            p_str = ""
            for p_var in p_key:
                p_str += "*" + p_var[1:]
            coe = self.coefficient[p_key]
            if not isinstance(coe, numbers.Number) and len(coe.coefficient) > 0:
                coe_str = "(" + str(coe) + ")"
            else:
                coe_str = str(coe)
            p_str = _check_unit(coe_str + p_str)

            if p_str[0] == "-":
                expression_str += p_str
            else:
                if expression_str == "":
                    expression_str += p_str
                else:
                    expression_str += "+" + p_str

        if not (is_zero(self.offset) and expression_str != ""):
            offset_str = str(self.offset)
            if offset_str[0] == "-":
                expression_str += offset_str
            else:
                if expression_str == "":
                    expression_str += offset_str
                else:
                    expression_str += "+" + offset_str
        return _check_unit(expression_str)

    def __radd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        q_ret = copy.copy(self)
        q_ret.clear()

        expr_add(self, other, q_ret)
        return q_ret

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):
        q_ret = copy.copy(self)
        q_ret.clear()

        expr_mul(self, other, q_ret)
        return q_ret

    def __truediv__(self, other):
        return self.__mul__(1 / other)

    def __pow__(self, other):
        q_ret = copy.copy(self)
        q_ret.clear()

        expr_pow(self, other, q_ret)
        return q_ret

    def __neg__(self):
        q_ret = copy.copy(self)
        q_ret.clear()

        expr_neg(self, q_ret)
        return q_ret

    def __eq__(self, other):
        return ConstraintDefinition(self - other, '==')

    def __ge__(self, other):
        return ConstraintDefinition(self - other, '>=')

    def __le__(self, other):
        return ConstraintDefinition(self - other, '<=')

    def __gt__(self, other):
        return ConstraintDefinition(self - other, '>')

    def __lt__(self, other):
        return ConstraintDefinition(self - other, '<')

    def get_variables(self):
        """获取变量名集合

            Returns:
                variables: (tuple) 返回构成expression的变量集合

        """
        variables = set()  # 放目标函数obj所包含的所有变量
        for key in self.coefficient:
            variables.add(key[0])
            if len(key) == 2:
                variables.add(key[1])

        return tuple(sorted(list(variables)))

    def get_max_deltas(self):
        """求出每个变量翻转引起目标函数变化的上界
        返回值negative_delta，positive_delta分别为该变量1->0和0->1所引起的最大变化量
        """
        obj_variables = self.get_variables()
        positive_delta = {var_x: 0 for var_x in obj_variables}
        negative_delta = {var_x: 0 for var_x in obj_variables}
        for var_tuple in self.coefficient.keys():
            for var_x in var_tuple:
                # 二次项受另一个变量影响，另一个变量为0时系数不引起变化
                if len(var_tuple) == 2:
                    positive_delta[var_x] += max(self.coefficient[var_tuple], 0)
                    # 变量0->1引起的变化量为-coefficient
                    negative_delta[var_x] += max(-self.coefficient[var_tuple], 0)

                # 一次项不受另一个变量影响，变量翻转一定引起变化
                elif len(var_tuple) == 1:
                    positive_delta[var_x] += self.coefficient[var_tuple]
                    # 变量0->1引起的变化为-coefficient,与二次项的情况一致
                    negative_delta[var_x] -= self.coefficient[var_tuple]
        return negative_delta, positive_delta

    def get_average_coefficient(self):
        """ 返回coefficient的平均值"""
        coe_sum = 0
        num_item = 0
        for _, coe in self.coefficient.items():
            coe_sum += abs(coe)
            num_item += 1
        return coe_sum / num_item



# -*- coding: utf-8 -*-
"""
模块: qubo

功能: QUBO建模工具
"""


from kaiwu_community.qubo._precision import check_qubo_matrix_bit_width, adjust_qubo_matrix_precision
from kaiwu_community.qubo._qubo_model import QuboModel, calculate_qubo_value, qubo_matrix_to_qubo_model
from kaiwu_community.qubo._error import QuboError


__all__ = [
    "check_qubo_matrix_bit_width",
    "adjust_qubo_matrix_precision",
    "QuboModel",
    "calculate_qubo_value",
    "qubo_matrix_to_qubo_model",
    "QuboError"
]



# -*- coding: utf-8 -*-
"""
模块: core.constraint

功能: 提供约束项基础定义类
"""
import math
import operator
import logging

from kaiwu_community.core._get_val import get_val
from kaiwu_community.core._error import KaiwuError

logger = logging.getLogger(__name__)


# Mapping from string to actual operator function
ops = {
    '==': operator.eq,
    '!=': operator.ne,
    '>': operator.gt,
    '>=': operator.ge,
    '<': operator.lt,
    '<=': operator.le,
}


class ConstraintDefinition:
    """约束定义

    Args:
        expr_left (Expression): 约束项左算子

        relation (string): 关系运算符

        expected_value(float): 约束项右算子， 缺省为0
    """

    def __init__(self, expr_left, relation, expected_value=0):
        self.left_operand = expr_left
        self.relation = relation
        self.expected_value = expected_value
        self.default_penalty = None

        # 是否需要二次方
        self.should_prepare = True

    def __str__(self):
        return f"{self.left_operand}{self.relation}{self.expected_value},"

    def __repr__(self) -> str:
        return self.__str__()

    def is_satisfied(self, solution_dict):
        """验证约束满足情况"""
        left = float(get_val(self.left_operand, solution_dict))
        right = float(self.expected_value)
        if self.relation != '==':
            return ops[self.relation](left, right)

        return abs(left - right) < 1e-5




def get_min_penalty_from_deltas(cons,
                                neg_delta, pos_delta, obj_vars, min_delta_method='diff'):
    """返回约束项cons对应的最小惩罚系数，惩罚项优先满足

    Args：
        cons：约束项的qubo表达式

        neg_delta: 各个变量1变为0时最大变化量的dict

        pos_delta: 各个变量0变为1时最大变化量的dict

        obj_vars: 第三个元素为变量列表

        min_delta_method: 声明在。分别用两种方法寻找最小变化值
                        MIN_DELTA_METHODS = {"diff": _get_constraint_min_deltas_diff,
                                     "exhaust": _get_constraint_min_deltas_exhaust}

    Examples:
        >>> import kaiwu_community as kw
        >>> x = [kw.core.Binary("b"+str(i)) for i in range(3)]
        >>> cons = kw.core.quicksum(x) - 1
        >>> obj = x[1]+2*x[2]
        >>> kw.core.get_min_penalty(obj, cons)
        2.0
    """
    constraint_var_quadratic = {}
    constraint_var_linear = {}
    for var_tuple, value in cons.coefficient.items():
        if len(var_tuple) == 2:
            for var_x in var_tuple:
                if var_x not in obj_vars:
                    continue
                if var_x not in constraint_var_quadratic:
                    constraint_var_quadratic[var_x] = [value]
                else:
                    constraint_var_quadratic[var_x].append(value)
        else:  # len == 1
            if var_tuple[0] in obj_vars:
                constraint_var_linear[var_tuple[0]] = value
    if min_delta_method not in MIN_DELTA_METHODS:
        raise KaiwuError("No such method for getting min delta")
    min_delta = MIN_DELTA_METHODS[min_delta_method](constraint_var_quadratic, constraint_var_linear)

    penalty = 0  # 惩罚系数，初始化
    for var_x, min_delta_x in min_delta.items():
        if var_x in pos_delta and min_delta_x != 0:
            penalty = max(pos_delta[var_x] / min_delta_x,
                          neg_delta[var_x] / min_delta_x, penalty)
    return penalty


def get_min_penalty_from_min_diff(cons, negative_delta, positive_delta):
    """ 根据objective的最大值，最小值估算约束项的最小惩罚系数

    Args:
        cons:  约束项

        negative_delta: objective最小值

        positive_delta: objective最大值

    Returns:
        找到的最小惩罚系数
    """
    # 如果正反向变化都没有，penalty置为1
    if not negative_delta or not positive_delta:
        return 1

    values = list(cons.coefficient.values())
    for i in range(len(cons.coefficient)):
        values.append(-values[i])
    values.append(0)
    values.sort()
    min_diff = math.inf
    for i in range(len(values) - 1):
        if values[i + 1] - values[i] > 0:
            min_diff = min(min_diff, values[i + 1] - values[i])
    penalty = 0  # 惩罚系数，初始化
    for max_delta_x in negative_delta.values():
        penalty = max(max_delta_x / min_diff, penalty)
    for max_delta_x in positive_delta.values():
        penalty = max(max_delta_x / min_diff, penalty)
    return penalty


def get_min_penalty_for_equal_constraint(obj, cons):
    """返回一次等式约束项cons对应的最小惩罚系数：把满足这个约束的解的某一位比特翻转一下的最坏情况。
    这个惩罚系数有效是指能够保证原问题的可行解是目标函数的局部最优（在一位比特翻转的局部意义下）。

    Args：
        obj: 原目标函数的qubo表达式。

        cons：线性的等式约束cons=0中的线性表达式。

    Returns:
        float: 一次等式约束项cons对应的最小惩罚系数.

    Examples：
        >>> import kaiwu_community as kw
        >>> x = [kw.core.Binary("b"+str(i)) for i in range(3)]
        >>> cons = kw.core.quicksum(x)-1
        >>> obj = x[1]+2*x[2]
        >>> kw.core.get_min_penalty_for_equal_constraint(obj,cons)
        2.0
    """
    # constraint_tuples为一次等式约束项cons所包含的所有变量及其相应的系数
    constraint_tuples = [(key[0], cons.coefficient[key]) for key in cons.coefficient]
    penalty = 0  # 惩罚系数，初始化
    neg_delta, pos_delta = obj.get_max_deltas()
    for vari, coef in constraint_tuples:
        if vari in neg_delta:
            penalty = max(pos_delta[vari] / pow(coef, 2), neg_delta[vari] / pow(coef, 2), penalty)
    return penalty


def _get_constraint_min_deltas_diff(constraint_var_quadratic, constraint_var_linear):
    """
    估计每个变量反转后在约束中引起的最小变化量
    准确求解需要枚举2^n种变量取值，用只有0、1、2个变量取1的情况来估计该值。
    """
    min_delta_dict = {}
    for var_x, value_list in constraint_var_quadratic.items():
        sorted_list = sorted(value_list)  # 对var_x相关的二次项进行排序
        if var_x in constraint_var_linear:
            # 有一次项，则翻转必然有影响，设为初值
            linear_coe = constraint_var_linear[var_x]
            min_delta = abs(linear_coe)
        else:
            # 没有一次项，初值为0
            linear_coe = 0
            min_delta = abs(sorted_list[-1])

        # 求出只考虑一个二次项时的最小变化量
        for i, val in enumerate(sorted_list):
            if linear_coe + val != 0:
                min_delta = min(min_delta, abs(linear_coe + val))
        i = 0
        j = len(sorted_list) - 1
        # 求出只考虑两个二次项时的最小变化量
        while i < j:
            new_delta = sorted_list[i] + sorted_list[j] + linear_coe
            if new_delta != 0:
                min_delta = min(min_delta, abs(new_delta))
            if -sorted_list[i] > sorted_list[j]:
                i += 1
            else:
                j -= 1
        min_delta_dict[var_x] = min_delta

    for var_x, value in constraint_var_linear.items():
        if var_x not in min_delta_dict:
            min_delta_dict[var_x] = value
    return min_delta_dict


def _get_constraint_min_deltas_exhaust(constraint_var_quadratic, constraint_var_linear):
    """
    计算每个变量反转后在约束中引起的最小变化量
    """
    min_delta_dict = {}
    for var_x, value_list in constraint_var_quadratic.items():
        if var_x in constraint_var_linear:
            # 有一次项，则翻转必然有影响，设为初值
            value_set = {constraint_var_linear[var_x]}
            min_delta = abs(constraint_var_linear[var_x])
        else:
            # 没有一次项，初值为0
            value_set = {0}
            min_delta = math.inf
        for coe in value_list:
            copy_set = value_set.copy()
            for sum_value in copy_set:
                value_set.add(sum_value + coe)
                if abs(sum_value + coe) < min_delta and sum_value + coe != 0:
                    min_delta = abs(sum_value + coe)
        min_delta_dict[var_x] = min_delta

    for var_x, value in constraint_var_linear.items():
        if var_x not in min_delta_dict:
            min_delta_dict[var_x] = value
    return min_delta_dict


MIN_DELTA_METHODS = {"diff": _get_constraint_min_deltas_diff,
                     "exhaust": _get_constraint_min_deltas_exhaust}


def get_soft_penalty(obj, cons):
    """返回soft约束项cons对应的惩罚系数，使其约束项和目标函数的系数平均数相同

    Args:
        obj (QuboExpression): 原目标函数的qubo表达式。

        cons (QuboExpression): 约束项的qubo表达式

    Returns:
        float: 返回约束项cons对应的惩罚系数.
    """
    avg_obj_coe = obj.get_average_coefficient()
    avg_cons_coe = cons.get_average_coefficient()
    return avg_obj_coe / avg_cons_coe


def get_min_penalty(obj, cons):
    """返回约束项cons对应的最小惩罚系数，惩罚项优先满足

    Args:
        obj: 原目标函数的qubo表达式。

        cons：约束项的qubo表达式

    Returns:
        float: 返回约束项cons对应的最小惩罚系数.

    Examples：
        >>> import kaiwu_community as kw
        >>> x = [kw.core.Binary("b"+str(i)) for i in range(3)]
        >>> cons = kw.core.quicksum(x) - 1
        >>> obj = x[1] + 2 * x[2]
        >>> kw.core.get_min_penalty(obj, cons)
        2.0
    """
    negative_delta, positive_delta = obj.get_max_deltas()
    return get_min_penalty_from_min_diff(cons, negative_delta, positive_delta)



"""
模块: qubo

功能: 提供QUBOModel相关的功能
"""
import logging
import numpy as np
from kaiwu_community.core import BinaryModel, ndarray, Binary, quicksum
from kaiwu_community.qubo._repr import _qubo_check, _qubo_details

logger = logging.getLogger(__name__)


class QuboModel(BinaryModel):
    """支持添加约束的QUBO模型类

    Args:
        objective (QuboExpression, optional): 目标函数. 默认为None
    """

    def __init__(self, objective=None):
        super().__init__(objective)
        self.qubo_expr_made = None

        self.made = False
        self.variables = None
        self.matrix = None

    def __repr__(self):
        return f"{self.__class__.__name__}({vars(self)!r})"

    def __str__(self):
        print_data = _qubo_details(self.objective)
        return print_data

    def invalidate_made_state(self):
        """Invalidate the made state when the model changes"""
        self.made = False
        self.qubo_expr_made = None
        self.matrix = None

    def make(self):
        """返回合并后的QUBO表达式

        Returns:
            BinaryExpression: 合并的约束表达式
        """
        if self.made:
            return self.qubo_expr_made

        constraint_list = self.get_constraints_expr_list()

        self.qubo_expr_made = self.objective + quicksum(constraint_list)

        variables_set = self.qubo_expr_made.get_variables()
        _qubo_check(variables_set, self.qubo_expr_made.coefficient)

        self.variables = dict(zip(variables_set, range(len(variables_set))))

        self.matrix = np.zeros((len(variables_set), len(variables_set)))
        for key in self.qubo_expr_made.coefficient:
            if len(key) == 2:
                self.matrix[self.variables[key[0]], self.variables[key[1]]] = \
                    self.qubo_expr_made.coefficient[key]
            else:
                self.matrix[self.variables[key[0]], self.variables[key[0]]] = \
                    self.qubo_expr_made.coefficient[key]
        self.made = True
        return self.qubo_expr_made

    def get_matrix(self):
        """获取QUBO矩阵

        Returns:
            numpy.ndarray: QUBO矩阵

        """
        self.compile_constraints()
        self.make()
        return self.matrix

    def get_variables(self):
        """获取qubo模型的variables"""
        self.compile_constraints()
        self.make()
        return self.variables

    def get_offset(self):
        """获取qubo模型的offset"""
        return self.objective.offset

    def get_sol_dict(self, qubo_solution):
        """根据解向量生成结果字典."""
        self.compile_constraints()
        self.make()
        return dict((k[1:], 1 if qubo_solution[idx] > 0 else 0)
                    for k, idx in self.variables.items() if k != '__spin__')


def calculate_qubo_value(qubo_matrix, offset, binary_configuration):
    """Q值计算器.

    Args:
        qubo_matrix (np.ndarray): QUBO矩阵.

        offset (float): 常数项

        binary_configuration (np.ndarray): 二进制配置

    Returns:
        output (float): Q值.

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> matrix = np.array([[1., 0., 0.],
        ...                    [0., 1., 0.],
        ...                    [0., 0., 1.]])
        >>> offset = 1.8
        >>> binary_configuration = np.array([0, 1, 0])
        >>> qubo_value = kw.qubo.calculate_qubo_value(matrix, offset, binary_configuration)
        >>> print(qubo_value)
        2.8
    """
    return (binary_configuration.dot(qubo_matrix)).dot(binary_configuration) + offset


def qubo_matrix_to_qubo_model(qubo_mat):
    """将qubo矩阵转化为qubo模型

    Args:
        qubo_mat (np.ndarray): QUBO矩阵

    Returns:
        QuboModel: QUBO模型

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> matrix = -np.array([[0, 8],
        ...                     [0, 0]])
        >>> kw.qubo.qubo_matrix_to_qubo_model(matrix).objective
        -8*b[0]*b[1]
    """
    vars_b = ndarray(len(qubo_mat), 'b', Binary)
    return QuboModel(vars_b.dot(qubo_mat).dot(vars_b))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# encoding=utf8
"""
SimpleSolver
"""
from kaiwu_community.core import get_sol_dict
from kaiwu_community.conversion import qubo_model_to_ising_model
from kaiwu_community.core import SolverBase
from kaiwu_community.common import hamiltonian


def _to_ising_matrix(qubo_model):
    ising_model = qubo_model_to_ising_model(qubo_model)
    ising_mat = ising_model.get_matrix()
    bias = ising_model.get_bias()
    vars_dict = ising_model.get_variables()
    return ising_mat, bias, vars_dict


class SimpleSolver(SolverBase):
    """实现用Optimizer直接对QuboModel进行求解

    Examples:
        >>> import kaiwu_community as kw
        >>> n = 10
        >>> W = 5
        >>> p = [i + 1 for i in range(n)]
        >>> w = [(i + 2) / 2 for i in range(n)]
        >>> x = kw.core.ndarray(n, 'x', kw.core.Binary)
        >>> qubo_model = kw.qubo.QuboModel()
        >>> qubo_model.set_objective(sum(x[i] * p[i] * (-1) for i in range(n)))
        >>> qubo_model.add_constraint(sum(x[i] * w[i] for i in range(n)) <= W, "c", penalty=10)
        >>> solver = kw.solver.SimpleSolver(kw.classical.BruteForceOptimizer())
        >>> sol_dict, qubo_val = solver.solve_qubo(qubo_model)
        >>> unsatisfied_count, result_dict = qubo_model.verify_constraint(sol_dict)
        >>> unsatisfied_count
        0

    Returns:
        tuple: Result dictionary and Result dictionary.

        - dict: Result dictionary. The key is the variable name, and the value is the corresponding spin value.

        - float: qubo value.
    """

    def solve_qubo(self, qubo_model):
        ising_mat, bias, vars_dict = _to_ising_matrix(qubo_model)
        solutions = self._optimizer.solve(ising_mat)
        if solutions is None:
            return None, None
        solution_dicts = get_sol_dict(solutions[0][:-1] * solutions[0][-1], vars_dict)
        return solution_dicts, hamiltonian(ising_mat, solutions)[0] + bias


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
一些简单的工具
"""

import numpy as np


def check_symmetric(mat, tolerance=1e-8):
    """检查矩阵是否为对称矩阵，允许一定误差

    Args:
        mat (np.ndarray): 矩阵.

        tolerance (float): 误差.

    Returns:
        bool: 是否为对称矩阵.

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> ising_matrix = -np.array([[ 0. ,  1. ,  0. ,  1. ,  1. ],
        ...                           [ 1. ,  0. ,  0. ,  1.,   1. ],
        ...                           [ 0. ,  0. ,  0. ,  1.,   1. ],
        ...                           [ 1. ,  1.,   1. ,  0. ,  1. ],
        ...                           [ 1. ,  1.,   1. ,  1. ,  0. ]])
        >>> print(kw.common.check_symmetric(ising_matrix))
        True
    """
    result = np.all(np.abs(mat - mat.T) < tolerance)
    return result


def hamiltonian(ising_matrix, c_list):
    """计算哈密顿量.

    Args:
        ising_matrix (np.ndarray): CIM Ising 矩阵.

        c_list (np.ndarray): 要计算哈密顿量的变量组合集合.

    Returns:
        np.ndarray: 哈密顿量集合.

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> ising_matrix = -np.array([[ 0. ,  1. ,  0. ,  1. ,  1. ],
        ...                     [ 1. ,  0. ,  0. ,  1.,   1. ],
        ...                     [ 0. ,  0. ,  0. ,  1.,   1. ],
        ...                     [ 1. ,  1.,   1. ,  0. ,  1. ],
        ...                     [ 1. ,  1.,   1. ,  1. ,  0. ]])
        >>> rng = np.random.default_rng(10)
        >>> optimizer = kw.classical.BruteForceOptimizer()
        >>> output = optimizer.solve(ising_matrix)
        >>> h = kw.common.hamiltonian(ising_matrix, output)
        >>> h   # doctest: +SKIP
        array([-0.60179257, -0.60179257, -0.60179257, -0.60179257, -0.60179257,
               -1.20358514, -0.60179257, -0.60179257, -0.60179257, -1.20358514])
    """
    # 方法1 by 王勇 邵帅 (最快版)
    return -np.einsum("ij,ij->i", (c_list.dot(ising_matrix)), c_list)

    # 方法2 by 王勇 邵帅
    # return ((c_list.dot(matrix))*c_list).sum(axis=1)
    #
    # 方法3 by 王勇
    # temp = c_list.dot(matrix)
    # return -np.array([temp[i] * c_list[i] for i in range(len(c_list))])
    #
    # 方法4 by 王勇
    # return -np.diag((c_list.dot(matrix)).dot(c_list.T)) # 空间复杂度高
    #
    # 方法5 by 邵帅
    # return -np.einsum('ij,ik,jk->i', c_list, c_list, matrix)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
模块: kaiwu_community

功能: 提供一系列CIM量子计算开发工具
"""

import logging
from kaiwu_community import qubo, core, ising, conversion, classical, solver

__version__ = "1.0.3"

logging.getLogger(__name__).addHandler(logging.NullHandler())
__all__ = ["qubo", "core","ising","conversion","classical","solver"]



# -*- coding: utf-8 -*-
"""
模块: core

功能: 基础类的定义

"""

from kaiwu_community.core._optimizer_base import OptimizerBase
from kaiwu_community.core._solver_base import SolverBase
from kaiwu_community.core._error import KaiwuError

from kaiwu_community.core._constraint import (ConstraintDefinition, get_min_penalty, get_min_penalty_from_min_diff,
                                              get_min_penalty_for_equal_constraint, get_min_penalty_from_deltas)
from kaiwu_community.core._penalty_method_constraint import PenaltyMethodConstraint
from kaiwu_community.core._get_val import get_sol_dict, get_array_val, get_val
from kaiwu_community.core._expression import update_constraint, Expression, expr_add, expr_mul, expr_neg, expr_pow
from kaiwu_community.core._binary_model import BinaryModel
from kaiwu_community.core._binary_expression import (BinaryExpression, Binary, quicksum, Placeholder, Integer)
from kaiwu_community.core._matrix import ndarray, dot, BinaryExpressionNDArray

__all__ = [
    "OptimizerBase",
    "SolverBase",
    "KaiwuError",
    "ConstraintDefinition","get_min_penalty","get_min_penalty_from_min_diff",
    "get_min_penalty_for_equal_constraint","get_min_penalty_from_deltas",
    "PenaltyMethodConstraint",
    "get_sol_dict",
    "get_array_val",
    "get_val",
    "update_constraint","Expression","expr_add","expr_mul","expr_neg","expr_pow",
    "BinaryModel",
    "BinaryExpression", "Binary", "quicksum", "Placeholder", "Integer",
    "ndarray", "dot", "BinaryExpressionNDArray"
]



# -*- coding: utf-8 -*-
"""
模块: qubo

功能: QUBO还原变量值
"""
import numbers
import numpy as np

# 构造结果字典
# 输入：结果向量，从ci.get_variables()得到的字典
def get_sol_dict(solution, vars_dict):
    """根据解向量和变量字典生成结果字典.

    Args:
        solution (np.ndarray): 解向量（spin）。

        vars_dict (dict): 变量字典，用cim_ising_model.get_variables()生成。

    Returns:
        dict: 结果字典。键为变量名，值为对应的spin值。

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> a = kw.core.Binary("a")
        >>> b = kw.core.Binary("b")
        >>> c = kw.core.Binary("c")
        >>> d = a + 2 * b + 4 * c
        >>> d = kw.qubo.QuboModel(d)
        >>> d_ising = kw.conversion.qubo_model_to_ising_model(d)
        >>> vars = d_ising.get_variables()
        >>> s = np.array([1, -1, 1])
        >>> kw.core.get_sol_dict(s, vars)
        {'a': np.float64(1.0), 'b': np.float64(0.0), 'c': np.float64(1.0)}
        """
    # 除了虚拟变量以外，遍历vars_dict字典，对每个键都找到其位置，并在bin_c中取值。
    return dict((k, (solution[vars_dict[k]] + 1) / 2)
                for k, _ in vars_dict.items() if k != '__spin__')


def get_val(qubo, sol_dict):
    """根据结果字典将spin值带入qubo变量.

    Args:
        qubo (QUBO表达式): QUBO表达式

        sol_dict (dict): 由get_sol_dict生成的结果字典。

    Returns:
        float: 带入qubo后所得的值

    Examples:
        >>> import kaiwu_community as kw
        >>> import numpy as np
        >>> a = kw.core.Binary("a")
        >>> b = kw.core.Binary("b")
        >>> c = kw.core.Binary("c")
        >>> d = a + 2 * b + 4 * c
        >>> qubo_model = kw.qubo.QuboModel(d)
        >>> d_ising = kw.conversion.qubo_model_to_ising_model(qubo_model)
        >>> ising_vars = d_ising.get_variables()
        >>> s = np.array([1, -1, 1])
        >>> sol_dict = kw.core.get_sol_dict(s, ising_vars)
        >>> kw.core.get_val(d, sol_dict)
        np.float64(5.0)
    """
    if isinstance(qubo, numbers.Number):
        return qubo
    value = qubo.offset  # 结果加上offset
    for k, val in qubo.coefficient.items():  # 遍历所有键值对
        for var in k:  # 对于每个键也遍历元组
            val *= sol_dict.get(var[1:], 0.0)
        value += val  # 累加一项的结果
    return value



def get_array_val(array, sol_dict):
    """根据结果字典将spin值带入qubo数组变量.

    Args:
        array (QUBOArray): QUBO数组

        sol_dict (dict): 由get_sol_dict生成的结果字典。

    Returns:
        np.ndarray: 带入qubo数组后所得的值数组

    Examples:
        >>> import kaiwu_community as kw
        >>> import numpy as np
        >>> x = kw.core.ndarray((2, 2), "x", kw.core.Binary)
        >>> y = x.sum()
        >>> y = kw.qubo.QuboModel(y)
        >>> y_ising = kw.conversion.qubo_model_to_ising_model(y)
        >>> ising_vars = y_ising.get_variables()
        >>> s = np.array([1, -1, 1, -1])
        >>> sol_dict = kw.core.get_sol_dict(s, ising_vars)
        >>> kw.core.get_array_val(x, sol_dict)
        array([[1., 0.],
               [1., 0.]])
    """
    val_array = np.zeros(array.shape)  # 创建与qubo数组相同形状的空数组
    for index in np.ndindex(array.shape):  # 遍历数组的所有下标
        val_array[index] = get_val(array[index], sol_dict)  # 每个下标都调用get_val
    return val_array


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
提供Ising矩阵精度校验相关功能
"""
import numpy as np


def calculate_ising_matrix_bit_width(ising_matrix, bit_width=8):
    """计算 ising 矩阵的参数位宽

    Args:
        ising_matrix (np.ndarray): ising 矩阵

        bit_width (int): 最大位宽限制

    Returns:
        dict: 返回Ising矩阵的精度和缩放因子

            - precision (int): Ising矩阵精度
            - multiplier (float): 缩放因子

    示例1:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> _matrix = -np.array([[ -0., 127., -12.,  -5.],
        ...                      [127.,  -0., -12., -12.],
        ...                      [-12., -12.,  -0.,  -9.],
        ...                      [ -5., -12.,  -9.,  -0.]])
        >>> kw.ising.calculate_ising_matrix_bit_width(_matrix)
        {'precision': 8, 'multiplier': np.float64(1.0)}

    示例2（缩放后符合要求）:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> _matrix = -np.array([[ -0., 12.7, -1.2,  -0.5],
        ...                      [12.7,  -0., -1.2, -1.2],
        ...                      [-1.2, -1.2, -0.,   -0.9],
        ...                      [-0.5, -1.2, -0.9,  -0.]])
        >>> kw.ising.calculate_ising_matrix_bit_width(_matrix)
        {'precision': 8, 'multiplier': np.float64(10.0)}

    示例3(缩放后也不符合要求):
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> _matrix = -np.array([[-488.,  516.,  -48.],
        ...                      [ 516., -516.,  -48.],
        ...                      [ -48.,  -48.,   60.]])
        >>> kw.ising.calculate_ising_matrix_bit_width(_matrix)
        {'precision': inf, 'multiplier': inf}
    """
    bit_width = bit_width - 1  # 留一位符号位

    if not np.any(ising_matrix):
        return np.inf, np.inf

    abs_matrix = np.abs(ising_matrix)
    threshold = 1e-10
    non_zero_elements = abs_matrix[np.where(np.abs(abs_matrix) > threshold)]
    normalization_factor = np.min(non_zero_elements)
    normalized_matrix = ising_matrix / normalization_factor
    abs_normalized_matrix = np.abs(normalized_matrix)
    scaling_factor_upper_limit = int(np.floor(2 ** bit_width / abs_normalized_matrix.max())) + 1

    for i in range(1, scaling_factor_upper_limit):
        scaled_normalized_matrix = ising_matrix * i / normalization_factor
        if np.array_equal(np.around(scaled_normalized_matrix, decimals=0),
                          np.around(scaled_normalized_matrix, decimals=8)):
            for j in range(1, bit_width + 1):
                upper_bound = 2 ** j
                if np.around(scaled_normalized_matrix.max()) == upper_bound:
                    continue
                if i < int(np.floor(upper_bound / abs_normalized_matrix.max())) + 1:
                    return {"precision": j + 1, "multiplier": i / normalization_factor}

    return {"precision": np.inf, "multiplier": np.inf}


def adjust_ising_matrix_precision(ising_matrix, bit_width=8):
    """调整 ising 矩阵精度, 通过此接口调整后矩阵可能会有较大的精度损失，比如矩阵有一个数远大于其它数时，调整后矩阵精度损失严重无法使用

    Args:
        ising_matrix(np.ndarray): 目标矩阵

        bit_width(int): 精度范围，目前只支持8位，有一位是符号位

    Returns:
        np.ndarray: 符合精度要求的 ising 矩阵

    Examples:
        >>> import numpy as np
        >>> import kaiwu_community as kw
        >>> ori_ising_mat1 = np.array([[0, 0.22, 0.198],
        ...                            [0.22, 0, 0.197],
        ...                            [0.198, 0.197, 0]])
        >>> ising_mat1 = kw.ising.adjust_ising_matrix_precision(ori_ising_mat1)
        >>> ising_mat1
        array([[  0, 127, 114],
               [127,   0, 114],
               [114, 114,   0]])
        >>> ori_ising_mat2 = np.array([[0, 0.22, 0.198],
        ...                            [0.22, 0, 50],
        ...                            [0.198, 50, 0]])
        >>> ising_mat2 = kw.ising.adjust_ising_matrix_precision(ori_ising_mat2)
        >>> ising_mat2  # The solutions obtained by qubo_mat2 and ori_qubo_mat2 matrices are quite different
        array([[  0,   1,   1],
               [  1,   0, 127],
               [  1, 127,   0]])
    """
    # 校验矩阵精度范围，如何符合要求直接返回
    result = calculate_ising_matrix_bit_width(ising_matrix, bit_width)
    if result.get('precision') > bit_width:
        # 保留一位符号位
        bit_width = bit_width - 1
        # 取绝对值
        abs_matrix = np.abs(ising_matrix)
        # 缩放因子
        factor = (2 ** bit_width - 1) / abs_matrix.max()
        # 小数位数
        decimal_bit_width = 0
        # 符合精度要求的ising矩阵
        ising_matrix = np.round(factor * ising_matrix * pow(2, decimal_bit_width)).astype(int)
    return ising_matrix


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
ising矩阵暴力求解器
"""
import math
import numpy as np
from kaiwu_community.core import OptimizerBase


class BruteForceOptimizer(OptimizerBase):
    """求解Ising模型矩阵的暴力求解器，慢而准."""

    def solve(self, ising_matrix=None):
        """求解Ising矩阵solve接口

        Args:
            ising_matrix (np.ndarray, optional): Ising矩阵. Defaults to None.

        Returns:
            np.ndarray: 1个或者多个能量最低的解向量.

        Examples:
            >>> import kaiwu_community as kw
            >>> import numpy as np
            >>> mat = np.array([[0, 2, -3],[2, 0, -1],[-3, -1, 0]])
            >>> optimizer = kw.classical.BruteForceOptimizer()
            >>> optimizer.solve(mat)
            array([[-1, -1,  1],
                   [-1, -1,  1]])

        """
        size = ising_matrix.shape[0]
        h_ret = math.inf
        solutions = []
        for i in range(2 ** size):
            val = i
            vlist = [0] * size
            for j in range(size):
                vlist[j] = val % 2
                val //= 2
            sol = np.array(vlist) * 2 - 1
            hmt = -sol.dot(ising_matrix).dot(sol)
            if hmt < h_ret:
                h_ret = hmt
                solutions = [sol]
            elif hmt == h_ret:
                solutions.append(sol)
        sols = np.array(solutions)
        sols[sols[:, -1] <= 0, :] *= -1
        return sols



# -*- coding: utf-8 -*-
"""
模块: qubo

功能: 提供QUBO相关类的字符串转化
"""
import numbers
import logging
from kaiwu_community.qubo._error import QuboError

logger = logging.getLogger(__name__)


def _qubo_check(variables_tuple, quadratic):
    if not variables_tuple:
        return
    var_type = variables_tuple[0][0]
    for var in variables_tuple:
        if var[0] != var_type:
            raise QuboError("Conflicting variable types, the QUBO expression contains "
                            "both \"binary\" and \"spin\".")
    for key in quadratic:
        if len(key) > 2:
            raise QuboError("Items higher than quadratic.")


def _get_placeholder(expr):
    placeholders_set = set()
    for value in expr.coefficient.values():
        if not isinstance(value, numbers.Number):
            for placeholder_tuple in value.coefficient:
                for var_p in placeholder_tuple:
                    placeholders_set.add(var_p)
    return tuple(sorted(list(placeholders_set)))


def _qubo_sketch(qubo_expr, q_check=True):
    variables_set = qubo_expr.get_variables()
    placeholders_set = _get_placeholder(qubo_expr)
    var_type = "unknown"
    if len(variables_set) == 0:
        logger.warning("Empty QUBO, the objective function is constant 0 (QUBO = 0).")
    else:
        var_type = variables_set[0][0]
        if q_check:
            _qubo_check(variables_set, qubo_expr.coefficient)
    return {"variable": variables_set, "placeholder": placeholders_set, "vtype": var_type}


def _expression_repr(expr):
    var_tuple_len = 0
    quadratic = {}
    print_data = ""
    for var_tuple, value in expr.coefficient.items():
        var_tuple_str = str(var_tuple)[1:-1]
        var_tuple_str = var_tuple_str.replace('\'b', '').replace('\'s', '').replace('\'', '')
        var_tuple_len = max(var_tuple_len, len(var_tuple_str))
        quadratic[var_tuple_str] = str(value)
    offset_str = str(expr.offset)

    sketch_dict = _qubo_sketch(expr)
    variables = sketch_dict["variable"]
    variables_str = ", ".join([var_x[1:] for var_x in variables])
    if sketch_dict["vtype"] == "s":
        var_type = "Spin):  "
    else:
        var_type = "Binary):"

    print_data += "  Variables(" + var_type
    print_data += variables_str + "\n"

    placeholders = sketch_dict["placeholder"]
    placeholders_str = ", ".join([var_x[1:] for var_x in placeholders])
    if placeholders_str != "":
        print_data += "  Placeholders:" + " "
        print_data += "    " + placeholders_str + "\n"

    print_data += "  QUBO offset:" + " "
    print_data += "     " + offset_str + "\n"

    print_data += "  QUBO coefficients:" + "\n"
    for key, value in quadratic.items():
        print_data += "    " + key.ljust(var_tuple_len + 1) + ": " + value + "\n"
    return print_data


def _qubo_details(expr):
    output_str = "QUBO Details:\n"
    output_str += _expression_repr(expr)
    return output_str
